{========================================================================}
{=                (c) 1995-97 SwiftSoft Ronald Dittrich                 =}
{========================================================================}
{=                          All Rights Reserved                         =}
{========================================================================}
{=  D 01099 Dresden             = Tel.: +0351-8012255                   =}
{=  Loewenstr.7a                = CIS : 100666,154, AOL: swiftsoft      =}
{========================================================================}
{=  Actual versions on http://members.aol.com/swiftsoft/mmtools.html    =}
{========================================================================}
{=  This code is for reference purposes only and may not be copied or   =}
{=  distributed in any format electronic or otherwise except one copy   =}
{=  for backup purposes.                                                =}
{=                                                                      =}
{=  No Delphi Component Kit or Component individually or in a collection=}
{=  subclassed or otherwise from the code in this unit, or associated   =}
{=  .pas, .dfm, .dcu, .asm or .obj files may be sold or distributed     =}
{=  without express permission from SwiftSoft.                          =}
{=                                                                      =}
{=  For more licence informations please refer to the associated        =}
{=  HelpFile.                                                           =}
{========================================================================}
{=  $Date: 16.11.1997 - 01:05:38 $                                      =}  
{========================================================================}
unit UExpert;

interface

{.$DEFINE _MMDEBUG}

uses
    Windows,
    SysUtils,
    Forms,
    Controls,
    Menus,
    Dialogs,
    Classes,
    Messages,
    Graphics,
    StdCtrls,
    Registry,
    ToolIntf,
    EditIntf,
    ExptIntf,
    VirtIntf
    {$IFDEF _MMDEBUG}
    ,MMDebug
    {$ENDIF}
    ;

type
    {-- TTimeStampExpert ------------------------------------------------------}
    TTimeStampExpert = class(TIExpert)
    private
       menEditExp    : TIMenuItemIntf;
       menEditSep    : TIMenuItemIntf;
       FExpKey       : string;
       FStrings      : TStringList;
       FEnabled      : Boolean;

       procedure     memHeaderClick(Sender: TIMenuItemIntf);
       procedure     InstallAddIn;
       procedure     GetRegInfo;
       procedure     InsertTimeStamp(const FileName: string);

    public
       constructor   Create; virtual;
       destructor    Destroy; override;

       function      GetName: string; override;
       function      GetStyle: TExpertStyle; override;
       function      GetState: TExpertState; override;
       function      GetIDString: string; override;
    end;

    {-- TProjectNotifier ------------------------------------------------------}
    TProjectNotifier = class(TIAddInNotifier)
    private
       FExpert       : TTimeStampExpert;
       FNotifiers    : TList;

       procedure     MakeModuleNotifier(const FileName: string);

    public
       constructor   Create(Expert: TTimeStampExpert);
       destructor    Destroy; override;

       procedure     FileNotification(NotifyCode: TFileNotification;
                                      const FileName: String; var Cancel: Boolean); override;

       property      Notifiers: TList read FNotifiers;
       property      Expert: TTimeStampExpert read FExpert;
    end;

    {-- TModuleNotifier -------------------------------------------------------}
    TModuleNotifier  = class(TIModuleNotifier)
    private
       FUnitFileName : string;
       FIntf         : TIModuleInterface;
       FProjNotifier : TProjectNotifier;

       procedure     GetFileNames;

    public
       constructor   Create(ModIntf: TIModuleInterface; ProjNotifier: TProjectNotifier);
       destructor    Destroy; override;

       procedure     Notify(NotifyCode: TNotifyCode); override;
       procedure     ComponentRenamed(ComponentHandle: Pointer;
                                      const OldName, NewName: string); override;

       property      ModIntf: TIModuleInterface read FIntf;
       property      UnitFileName: string read FUnitFileName;
       property      ProjNotifier: TProjectNotifier read FProjNotifier;
    end;

function InitExpert(ToolServices: TIToolServices;
                    Registerproc: TExpertRegisterproc;
                    var Terminate: TExpertTerminateProc): Boolean; export; stdcall;

var
   expEdit : TTimeStampExpert = nil;
   Notifier: TProjectNotifier = nil;

implementation

uses Dialog;

{==============================================================================}
procedure DebugString(const Msg: string);
begin
   {$IFDEF _MMDEBUG}
   DB_WriteStrLn(0, Msg);
   {$ENDIF}
end;

{=========================================================================}
procedure SaveInRegistry(_RootKey:HKEY;_Localkey,_Field:String;Value:Variant);
begin
   with TRegistry.Create do
   try
      { default is  RootKey=HKEY_CURRENT_USER  }
      case _RootKey of
         HKEY_CLASSES_ROOT,
         HKEY_CURRENT_USER,
         HKEY_LOCAL_MACHINE,
         HKEY_USERS,
         HKEY_PERFORMANCE_DATA,
         HKEY_CURRENT_CONFIG,
         HKEY_DYN_DATA : RootKey := _RootKey;
      end;

      OpenKey(_Localkey,True);
      case VarType(Value) of
         varByte,
         varNull,
         varInteger,
         varSmallint: WriteInteger (_Field,Value);
         varSingle,
         varDouble  : WriteFloat   (_Field,Value);
         varCurrency: WriteCurrency(_Field,Value);
         varDate    : WriteDateTime(_Field,Value);
         varBoolean : WriteBool    (_Field,Value);
         varString  : WriteString  (_Field,Value);
      end;
      CloseKey;

   finally
      Free;
   end;
end;

{=========================================================================}
function GetFromRegistry(_RootKey:HKEY;_Localkey,_Field:String;Value:Variant): Variant;
begin
   Result := Value;
   with TRegistry.Create do
   try
      { default is  RootKey=HKEY_CURRENT_USER  }
      case _RootKey of
        HKEY_CLASSES_ROOT,
        HKEY_CURRENT_USER,
        HKEY_LOCAL_MACHINE,
        HKEY_USERS,
        HKEY_PERFORMANCE_DATA,
        HKEY_CURRENT_CONFIG,
        HKEY_DYN_DATA : RootKey := _RootKey;
      end;

      if OpenKey(_Localkey, False) then
      begin
         if ValueExists(_Field) then
         case VarType(Value) of
              varByte,
              varNull,
              varInteger,
              varSmallint: Result := ReadInteger(_Field);
              varSingle  : Result := ReadFloat   (_Field);
              varDouble  : Result := ReadFloat   (_Field);
              varCurrency: Result := ReadCurrency(_Field);
              varDate    : Result := ReadDateTime(_Field);
              varBoolean : Result := ReadBool    (_Field);
              varString  : Result := ReadString  (_Field);
         end;
         CloseKey;
      end;

   finally
      Free;
   end;
end;

{== TTimeStampExpert ==========================================================}
constructor TTimeStampExpert.Create;
begin
   inherited Create;

   FEnabled := True;
   FStrings := TStringList.Create;
   FExpKey  := 'Software\SwiftSoft\TimeStamp';

   InstallAddIn;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
destructor TTimeStampExpert.Destroy;
begin
   menEditExp.DestroyMenuItem;
   menEditSep.DestroyMenuItem;

   inherited Destroy;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
procedure TTimeStampExpert.InstallAddIn;
var
   menMain    : TIMainMenuIntf;
   menEdit    : TIMenuItemIntf;
   menEditMenu: TIMenuItemIntf;
   idx        : integer;

begin
   if (menEditExp <> nil) then exit;

   GetRegInfo;

   Notifier := TProjectNotifier.Create(Self);

   menMain := ToolServices.GetMainMenu;
   if Assigned(menMain) then
   try
      { Find parent menu }
      menEdit := menMain.FindMenuItem('EditMenu');
      if Assigned(menEdit) then
      try
         menEditMenu := menMain.FindMenuItem('EditObjectItem');
         if not assigned(menEditMenu) or (menEditMenu.GetName <> 'EditObjectItem') then
            menEditMenu := menMain.FindMenuItem('EditAddToInterfaceItem');

         if Assigned(menEditMenu) then
         begin
            idx := menEditMenu.GetIndex+1;
            menEditMenu.DestroyMenuItem;
         end
         else idx := 0;

         menEditSep := menEdit.InsertItem(idx, '-', 'TimeExpSepa1', '', 0, 0, 0, [mfVisible], NIL);
         if assigned(menEditSep) then
            menEditExp := menEdit.InsertItem(menEditSep.GetIndex+1,
                                             'T&ime Stamp...',
                                             'TimeExpTimeStamp', '',
                                              0, 0, 0,
                                              [mfEnabled, mfVisible], memHeaderClick);

      finally
         menEdit.DestroyMenuItem;
      end;

   finally
      menMain.Free;
   end;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
procedure TTimeStampExpert.GetRegInfo;
var
   i,nLines: integer;
   s: string;

begin
   FStrings.Clear;
   FEnabled := GetFromRegistry(HKEY_CURRENT_USER,
                               FExpKey,'Enabled',True);
   nLines   := GetFromRegistry(HKEY_CURRENT_USER,
                               FExpKey,'Lines',0);
   for i := 0 to nLines-1 do
   begin
      s := GetFromRegistry(HKEY_CURRENT_USER,
                           FExpKey,'Line'+IntToStr(i),'');
      if (s <> '') then FStrings.Add(s);
   end;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
procedure TTimeStampExpert.memHeaderClick(Sender: TIMenuItemIntf);
var
   i: integer;

begin
   with TExpertDialog.Create(nil) do
   try
      RichEdit.Lines.Assign(FStrings);

      ckbEnabled.Checked := FEnabled;

      if (ShowModal = mrOK) then
      with RichEdit do
      begin
         SaveInRegistry(HKEY_CURRENT_USER,
                        FExpKey,'Enabled',ckbEnabled.Checked);

         SaveInRegistry(HKEY_CURRENT_USER,
                        FExpKey,'Lines',Lines.Count);
         for i := 0 to RichEdit.Lines.Count-1 do
         begin
            SaveInRegistry(HKEY_CURRENT_USER,
                           FExpKey,'Line'+IntToStr(i),Lines[i]);
         end;
         GetRegInfo;
      end;

   finally
      Free;
   end;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
procedure TTimeStampExpert.InsertTimeStamp(const FileName: string);
const
   cr = chr(13)+chr(10);
var
   mModule: TIModuleInterface;
   mEdit  : TIEditorInterface;
   reader : TIEditReader;
   writer : TIEditWriter;
   oldPos : integer;
   readPos: integer;
   s,sz   : string;
   HW     : HWND;
   Marker : string;
   Temp   : string;
   i,j    : integer;
   EOF    : Boolean;

   function ReadEditorLine : String;                      // Read a return a line from the Editor
   var
      CharPos: Integer;
   begin
      SetLength(Result,1024);                             // Longest Line !!!!
      CharPos := 1;                                       // begin at the beginning
      repeat
        try
           EOF := reader.GetText(readPos,
                                 @Result[CharPos],1) <> 1;// Read a character
        except                                            // Exception trap
          begin
             Eof := True;                                 // Eof JIC
             Result := '';                                // No Result
          end;
        end;
        Inc(readPos);                                     // Read one so increment the buffer position
        if Result[CharPos] = #10 then
        begin                                             // was it a LineFeed and if so...
           Result[CharPos] := #0;                         // Null the last char
           SetLength(Result,CharPos - 1);                 // set the length
           Exit;                                          // bye bye
        end;
        Inc(CharPos);

      until EOF;                                          // repeat till end of file
      Result := '';                                       // if we get here we are eof so ''
   end;


begin
   if (Toolservices = nil) or not FEnabled or (FStrings.Count = 0) then exit;

   Marker := UpperCase(ExtractFileExt(FileName));

   if (Marker = '.PAS') then
   begin
      Marker := 'UNIT ';
   end
   else if (Marker = '.DPR') then
   begin
      Marker := 'LIBRARY ';
   end
   else exit;

   Temp := UpperCase(ExtractFileName(FileName));

   i := Pos('.',Temp);
   if (i <> 0) then
      Temp := Copy(Temp, 1, i-1)+';'
   else
      Temp := '';

   marker := Marker + Temp;

   DebugString('Marker: '+Marker);

   if not Toolservices.IsFileOpen(FileName) then
      if not Toolservices.OpenFile(FileName) then exit;

   HW := FindWindow('TeditWindow',nil);
   if HW > 0 then
   try
      mModule := ToolServices.GetModuleInterface(FileName);
      mEdit   := mModule.GetEditorInterface;

      reader  := mEdit.CreateReader;
      try
         readPos := 0;
         repeat
            oldPos := readPos;
            Temp := ReadEditorLine;
            DebugString(Temp);
         until EOF or (Pos(Marker,UpperCase(Temp)) > 0);

      finally
         reader.Free;
      end;

      writer  := mEdit.CreateWriter;
      try
         if not EOF then
         begin
            writer.DeleteTo(oldPos);
         end;

         for i := 0 to FStrings.Count-1 do
         begin
            j := Pos('$Date:',FStrings[i]);
            if (j <> 0) then
            begin
               s := FStrings[i];
               sz := Copy(s,1,j)+'Date: '+DateToStr(date)+' - '+TimeToStr(Time);
               Delete(s,1,j);
               j := Pos('$',s);
               if (j <> 0) then
               begin
                  sz := sz+' '+Copy(s,j,Length(s));
               end;
               writer.Insert(PChar(sz+cr));
            end
            else writer.Insert(PChar(FStrings[i]+cr));
         end;

      finally
         writer.free;
      end;

   finally
      mEdit.Free;
      mModule.Free;
   end;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
function TTimeStampExpert.GetName: String;
begin
   Result := 'TimeStamp Expert';
end;

{-- TTimeStampExpert ----------------------------------------------------------}
function TTimeStampExpert.GetStyle: TExpertStyle;
begin
   Result := esAddIn;
end;

{-- TTimeStampExpert ----------------------------------------------------------}
function TTimeStampExpert.GetState: TExpertState;
begin
   Result := [esEnabled]
end;

{-- TTimeStampExpert ----------------------------------------------------------}
function TTimeStampExpert.GetIDString: String;
begin
  Result := 'SwiftSoft.TimeStampExpert';
end;

{== TProjectNotifier ==========================================================}
constructor TProjectNotifier.Create(Expert: TTimeStampExpert);
begin
   FExpert    := Expert;
   FNotifiers := TList.Create;

   inherited Create;

   ToolServices.AddNotifier(Self);
end;

{-- TProjectNotifier ----------------------------------------------------------}
destructor TProjectNotifier.Destroy;
var
   i: integer;
begin
   for i := 0 to FNotifiers.Count-1 do
   begin
      TModuleNotifier(FNotifiers[i]).Free;
   end;
   FNotifiers.Free;

   ToolServices.RemoveNotifier(Self);

   inherited Destroy;
end;

{-- TProjectNotifier ----------------------------------------------------------}
procedure TProjectNotifier.FileNotification(NotifyCode: TFileNotification;
                                            const FileName: String; var Cancel: Boolean);
begin
   case NotifyCode of
      fnProjectOpened,
      fnFileOpened,
      fnAddedToProject: if (FileName <> '') then
      begin
         MakeModuleNotifier(FileName);
         DebugString('Opened: '+Filename);
      end;
   end;
end;

{-- TProjectNotifier ----------------------------------------------------------}
procedure TProjectNotifier.MakeModuleNotifier(const FileName: string);
var
   ModIntf: TIModuleInterface;

begin
   ModIntf := ToolServices.GetModuleInterface(FileName);
   if (ModIntf <> nil) then
   begin
      Notifiers.Add(TModuleNotifier.Create(ModIntf,Self));
   end;
end;

{== TModuleNotifier ===========================================================}
constructor TModuleNotifier.Create(ModIntf: TIModuleInterface; ProjNotifier: TProjectNotifier);
begin
   FIntf         := ModIntf;
   FProjNotifier := ProjNotifier;

   inherited Create;

   FIntf.AddNotifier(Self);

   GetFileNames;
end;

{-- TModuleNotifier -----------------------------------------------------------}
destructor TModuleNotifier.Destroy;
begin
   if (ModIntf <> nil) then
   begin
       ModIntf.RemoveNotifier(Self);
       ModIntf.Free;
   end;

   inherited Destroy;
end;

{-- TModuleNotifier -----------------------------------------------------------}
procedure TModuleNotifier.GetFileNames;
var
   EditIntf: TIEditorInterface;

begin
  EditIntf := ModIntf.GetEditorInterface;
  if (EditIntf = nil) then
      FUnitFilename := ''
  else
  begin
      FUnitFileName := EditIntf.FileName;
      EditIntf.Free;
  end;
end;

{-- TModuleNotifier -----------------------------------------------------------}
procedure TModuleNotifier.Notify(NotifyCode: TNotifyCode);
begin
   case NotifyCode of
        ncModuleDeleted: begin
                            DebugString('Removed: '+FUnitFilename);
                            ProjNotifier.Notifiers.Remove(Self);
                            Free;
                         end;
        ncModuleRenamed: begin
                            GetFileNames;
                            DebugString('Renamed: '+FUnitFilename);
                         end;
        ncBeforeSave   : begin
                            expEdit.InsertTimeStamp(FUnitFilename);
                            DebugString('Saved: '+FUnitFilename);
                         end;
   end;
end;

{-- TModuleNotifier -----------------------------------------------------------}
procedure TModuleNotifier.ComponentRenamed(ComponentHandle: Pointer;
                                           const OldName, NewName: string);
begin
   ;
end;

{------------------------------------------------------------------------------}
function InitExpert(ToolServices: TIToolServices; Registerproc: TExpertRegisterproc;
                    var Terminate: TExpertTerminateProc): Boolean; export; stdcall;
begin
   Result := ExptIntf.ToolServices = nil;
   if not Result then exit;

   ExptIntf.ToolServices := ToolServices;
   if Assigned(ToolServices) then
      Application.Handle := ToolServices.GetParentHandle;

   Terminate := nil;

   expEdit := TTimeStampExpert.Create;
   RegisterProc(expEdit);
end;

end.
